import * as url from 'url';
import * as _ from 'lodash';
import * as Electron from 'electron';

import {
  MultiWindowChannel,
  MultiWindowWindowChannel,
} from '../types/channel.type';
import {head} from 'lodash';

export interface Template {
  filePath?: string;
  html?: string;
}

export interface MultiWindowPositionAlign {
  vertical?: 'top' | 'middle' | 'bottom';
  horizontal?: 'left' | 'center' | 'right';
}

export interface MultiWindowPosition {
  displayTarget?: number;
  size?: 'full' | number | {width?: 'full' | number; height?: 'full' | number};
  point?: Electron.Point;
  align?: MultiWindowPositionAlign;
}

export interface DefaultOptions {
  appIcon?: string;
  optionsForBrowserWindow?: Electron.BrowserWindowConstructorOptions;
  template: Template;
  position?: MultiWindowPosition;
}

export const defaultOptions: DefaultOptions = {
  optionsForBrowserWindow: {
    alwaysOnTop: true,
    skipTaskbar: true,
    resizable: false,
    show: false,
    frame: false,
    transparent: true,
    acceptFirstMouse: true,
  },
  template: {
    html: `
<html>\n
<head></head>\n
<body style="overflow: hidden; -webkit-user-select: none;">\n
  <div id="container">\n
    <img src="" id="appIcon" />\n
    <img src="" id="image" />\n
    <div id="text">\n
      <b id="title"></b>\n
      <p id="message"></p>\n
    </div>\n
    <div id="close">X</div>\n
  </div>\n
</body>\n
</html>
    `,
  },
};

export type OnShow4NotificationWindow = () => Promise<void>;
export type OnClose4NotificationWindow = () => Promise<void>;
export type OnMessage4NotificationWindow = (
  channel: string,
  ...args: any[]
) => Promise<void>;

export interface MultiWindow {
  id?: number;
  optionsForBrowserWindow?: Electron.BrowserWindowConstructorOptions;
  template: Template;
  position?: MultiWindowPosition;
  data?: any;
  onShow?: OnShow4NotificationWindow;
  onClose?: OnClose4NotificationWindow;
  onMessage?: OnMessage4NotificationWindow;
}

type Activated = {
  multiWindow: MultiWindow;
  browserWindow: Electron.BrowserWindow;
};

export class MultiWindowService {
  private readonly __defaultOptions: DefaultOptions;
  private __multiWindowId = 0;
  private __activated: Activated[] = [];

  constructor(options?: DefaultOptions) {
    this.__defaultOptions = _.assignIn(defaultOptions, options);
  }

  init(): void {
    this.__initListeners();
  }

  destroy(): void {
    this.__destroyListeners();
  }

  open(multiWindow: MultiWindow): Promise<void> {
    multiWindow.id = this.__nextMultiWindowId();
    return new Promise<void>((resolve, reject) => {
      this.__show(multiWindow)
        .then(() => {
          resolve();
        })
        .catch(reason => {
          reject(reason);
        });
    });
  }

  close(id: number): void {
    const activated = this.__findActivatedByMultiWindowId(id);
    if (!activated) {
      return;
    }

    activated.browserWindow.close();
  }

  message(id: number, channel: string, ...args: any[]): void {
    const activated = this.__findActivatedByMultiWindowId(id);
    if (!activated) {
      return;
    }
    activated.browserWindow.webContents.send(channel, ...args);
  }

  private __show(
    multiWindow: MultiWindow
  ): Promise<Electron.BrowserWindow | undefined> {
    const __this = this;
    return new Promise<Electron.BrowserWindow | undefined>(
      (resolve, reject) => {
        const dimension = __this.__dimension(multiWindow);
        __this
          .__getWindow(multiWindow)
          .then(browserWindow => {
            browserWindow.setPosition(dimension.point.x, dimension.point.y);
            __this.__activated.push({
              multiWindow,
              browserWindow,
            });

            if (!!multiWindow.onShow) {
              multiWindow.onShow();
            }

            browserWindow.webContents.send(MultiWindowChannel.ready4Window);

            browserWindow.showInactive();
            return resolve(browserWindow);
          })
          .catch(reason => {
            reject(reason);
          });
      }
    );
  }

  private __getWindow(
    multiWindow: MultiWindow
  ): Promise<Electron.BrowserWindow> {
    const __this = this;
    return new Promise<Electron.BrowserWindow>((resolve, reject) => {
      const constructorOptions = _.assignIn(
        this.__defaultOptions.optionsForBrowserWindow,
        multiWindow.optionsForBrowserWindow
      );

      const browserWindow = new Electron.BrowserWindow(constructorOptions);
      browserWindow.setVisibleOnAllWorkspaces(true);
      browserWindow.loadURL(__this.__templateUrl(multiWindow));
      browserWindow.webContents.once('did-finish-load', () => {
        browserWindow.webContents.send(
          MultiWindowChannel.init4Window,
          multiWindow.data
        );
        resolve(browserWindow);
      });
      browserWindow.once('closed', () => {
        __this.__onClosed4MultiWindow(browserWindow);
      });
    });
  }

  private __dimension(
    multiWindow: MultiWindow
  ): {point: Electron.Point; size: Electron.Size} {
    if (
      (!this.__defaultOptions.optionsForBrowserWindow?.width ||
        !this.__defaultOptions.optionsForBrowserWindow?.height) &&
      (!multiWindow.optionsForBrowserWindow?.width ||
        !multiWindow.optionsForBrowserWindow?.height)
    ) {
      throw Error('There is not enough size information.');
    }

    if (
      (!this.__defaultOptions.position && !multiWindow.position) ||
      (!this.__defaultOptions.position?.point &&
        !this.__defaultOptions.position?.align) ||
      (!multiWindow.position?.point && !multiWindow.position?.align)
    ) {
      throw Error('There is not enough location information.');
    }

    const displayTarget = !!multiWindow.position?.displayTarget
      ? multiWindow.position?.displayTarget
      : !!this.__defaultOptions.position?.displayTarget
      ? this.__defaultOptions.position.displayTarget
      : 0;

    const displays = Electron.screen.getAllDisplays();

    let _displayTarget = displayTarget;

    if (displays.length - 1 < _displayTarget) {
      _displayTarget = 0;
    }

    const display = displays[_displayTarget];

    let size: Electron.Size = {width: 0, height: 0};

    if (!!multiWindow.position.size) {
      const percentage = (value: number, p: number): number => {
        let _p = p;
        if (0 > p) {
          _p = 0;
        }
        if (100 < p) {
          _p = 100;
        }

        return value * (p / 100);
      };

      if ('full' === multiWindow.position.size) {
        size = display.workAreaSize;
      } else if (_.isNumber(multiWindow.position.size)) {
        size = {
          width: percentage(
            display.workAreaSize.width,
            multiWindow.position.size
          ),
          height: percentage(
            display.workAreaSize.height,
            multiWindow.position.size
          ),
        };
      } else {
        let width = 0;
        let height = 0;
        if ('full' === multiWindow.position.size.width) {
          width = display.workAreaSize.width;
        } else {
          width = percentage(
            display.workAreaSize.width,
            multiWindow.position.size.width as number
          );
        }
        if ('full' === multiWindow.position.size.height) {
          height = display.workAreaSize.height;
        } else {
          height = percentage(
            display.workAreaSize.height,
            multiWindow.position.size.height as number
          );
        }
        size = {width, height};
      }
    } else {
      if (
        !!multiWindow.optionsForBrowserWindow?.width &&
        !!multiWindow.optionsForBrowserWindow?.height
      ) {
        size = {
          width: multiWindow.optionsForBrowserWindow?.width,
          height: multiWindow.optionsForBrowserWindow?.height,
        };
      } else if (
        !!this.__defaultOptions.optionsForBrowserWindow?.width &&
        !!this.__defaultOptions.optionsForBrowserWindow?.height
      ) {
        size = {
          width: this.__defaultOptions.optionsForBrowserWindow?.width,
          height: this.__defaultOptions.optionsForBrowserWindow?.height,
        };
      } else {
        size = {width: 0, height: 0};
      }
    }

    const point: Electron.Point | undefined = !!multiWindow.position?.point
      ? multiWindow.position?.point
      : !!this.__defaultOptions.position.point
      ? this.__defaultOptions.position.point
      : undefined;

    if (!!point) {
      return {
        point,
        size,
      };
    }

    const align: MultiWindowPositionAlign | undefined = !!multiWindow.position
      ?.align
      ? multiWindow.position?.align
      : !!this.__defaultOptions.position.align
      ? this.__defaultOptions.position.align
      : undefined;

    if (!!align) {
      let x = 0;
      let y = 0;

      switch (align.horizontal) {
        case 'center':
          {
            x =
              display.workArea.x +
              (display.workAreaSize.width - size.width) / 2;
          }
          break;
        case 'right':
          {
            x = display.workArea.x + (display.workAreaSize.width - size.width);
          }
          break;
        default:
          {
            x = display.workArea.x;
          }
          break;
      }

      switch (align.vertical) {
        case 'middle':
          {
            y =
              display.workArea.y +
              (display.workAreaSize.height - size.height) / 2;
          }
          break;
        case 'bottom':
          {
            y =
              display.workArea.y + (display.workAreaSize.height - size.height);
          }
          break;
        default:
          {
            y = display.workArea.y;
          }
          break;
      }

      return {
        point: {
          x,
          y,
        },
        size,
      };
    }

    throw Error('There is not enough position information.');
  }

  private __nextMultiWindowId(): number {
    this.__multiWindowId++;

    if (Number.MAX_SAFE_INTEGER < this.__multiWindowId) {
      this.__multiWindowId = 0;
    }

    return this.__multiWindowId;
  }

  private __findActivatedByMultiWindowId(id: number): Activated | undefined {
    const i = this.__activated.findIndex(a => a.multiWindow.id === id);
    if (-1 === i) {
      return undefined;
    }
    return this.__activated[i];
  }

  private __findActivatedByBrowserWindowId(id: number): Activated | undefined {
    const i = this.__activated.findIndex(a => a.browserWindow.id === id);
    if (-1 === i) {
      return undefined;
    }
    return this.__activated[i];
  }

  private __templateUrl(multiWindow: MultiWindow): string {
    let u: string = '';

    const filePath2Url = (filePath: string) => {
      return url.format({
        protocol: 'file:',
        pathname: filePath,
        slashes: true,
      });
    };

    const html2Url = (html: string) => {
      return 'data:text/html,' + encodeURIComponent(html);
    };

    if (!!multiWindow.template.filePath) {
      return filePath2Url(multiWindow.template.filePath);
    }

    if (!!multiWindow.template.html) {
      return html2Url(multiWindow.template.html);
    }

    if (!!this.__defaultOptions.template.filePath) {
      return filePath2Url(this.__defaultOptions.template.filePath);
    }

    if (!!this.__defaultOptions.template.html) {
      return html2Url(this.__defaultOptions.template.html);
    }

    return !!defaultOptions.template.html ? defaultOptions.template.html : '';
  }

  private __rePositioning(): void {
    this.__activated.forEach(a => {});
  }

  private __initListeners(): void {
    Electron.ipcMain.on(
      MultiWindowWindowChannel.onClose,
      this.__onClose4NotificationWindow.bind(this)
    );
    Electron.ipcMain.on(
      MultiWindowWindowChannel.onMessage,
      this.__onMessage4NotificationWindow.bind(this)
    );

    Electron.screen.on(
      'display-added',
      this.__onDisplayAdded4ElectronScreen.bind(this)
    );
    Electron.screen.on(
      'display-removed',
      this.__onDisplayRemoved4ElectronScreen.bind(this)
    );
    Electron.screen.on(
      'display-metrics-changed',
      this.__onDisplayMetricsChanged4ElectronScreen.bind(this)
    );
  }

  private __destroyListeners(): void {
    Electron.ipcMain.off(
      MultiWindowWindowChannel.onClose,
      this.__onClose4NotificationWindow.bind(this)
    );
    Electron.ipcMain.off(
      MultiWindowWindowChannel.onMessage,
      this.__onMessage4NotificationWindow.bind(this)
    );

    Electron.screen.off(
      'display-added',
      this.__onDisplayAdded4ElectronScreen.bind(this)
    );
    Electron.screen.off(
      'display-removed',
      this.__onDisplayRemoved4ElectronScreen.bind(this)
    );
    Electron.screen.off(
      'display-metrics-changed',
      this.__onDisplayMetricsChanged4ElectronScreen.bind(this)
    );
  }

  private __onClose4NotificationWindow(
    event: Electron.IpcMainEvent,
    notificationId: number
  ): void {
    const activated = this.__findActivatedByMultiWindowId(notificationId);
    if (!activated || !activated.multiWindow.onClose) {
      return;
    }

    activated.multiWindow.onClose();
  }

  private __onMessage4NotificationWindow(
    event: Electron.IpcMainEvent,
    notificationId: number,
    channel: string,
    ...args: any[]
  ): void {
    const activated = this.__findActivatedByMultiWindowId(notificationId);
    if (!activated || !activated.multiWindow.onMessage) {
      return;
    }

    activated.multiWindow.onMessage(channel, ...args);
  }

  private __onDisplayAdded4ElectronScreen(
    event: Electron.Event,
    newDisplay: Electron.Display
  ): void {
    this.__rePositioning();
  }

  private __onDisplayRemoved4ElectronScreen(
    event: Electron.Event,
    oldDisplay: Electron.Display
  ): void {
    this.__rePositioning();
  }

  private __onDisplayMetricsChanged4ElectronScreen(
    event: Electron.Event,
    display: Electron.Display,
    changedMetrics: string[] // bounds, workArea, scaleFactor, rotation
  ): void {
    this.__rePositioning();
  }

  private __onClosed4MultiWindow(browserWindow: Electron.BrowserWindow): void {
    const activated = this.__findActivatedByBrowserWindowId(browserWindow.id);
    if (!activated) {
      return;
    }

    const i = this.__activated.findIndex(
      a => a.browserWindow.id === browserWindow.id
    );
    if (-1 === i) {
      return;
    }

    this.__activated.splice(i, 1);
  }
}
